////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2015 Saxonica Limited.
// This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
// If a copy of the MPL was not distributed with this file, You can obtain one at http://mozilla.org/MPL/2.0/.
// This Source Code Form is "Incompatible With Secondary Licenses", as defined by the Mozilla Public License, v. 2.0.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

package com.saxonica.testdriver;


import com.saxonica.config.ProfessionalConfiguration;
import com.saxonica.ee.pattern.PatternOptimizationEE;
import com.saxonica.ee.stream.Posture;
import com.saxonica.ee.stream.PostureAndSweep;
import com.saxonica.ee.stream.Sweep;
import net.sf.saxon.expr.Expression;
import net.sf.saxon.expr.Literal;
import net.sf.saxon.expr.parser.ContextItemStaticInfo;
import net.sf.saxon.lib.FeatureKeys;
import net.sf.saxon.om.TreeModel;
import net.sf.saxon.option.axiom.AxiomObjectModel;
import net.sf.saxon.option.dom4j.DOM4JObjectModel;
import net.sf.saxon.option.jdom.JDOMObjectModel;
import net.sf.saxon.option.jdom2.JDOM2ObjectModel;
import net.sf.saxon.option.xom.XOMObjectModel;
import net.sf.saxon.s9api.*;
import net.sf.saxon.testdriver.Environment;
import net.sf.saxon.testdriver.Xslt30TestSuiteDriverHE;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.AnyItemType;

import java.util.ArrayList;
import java.util.List;

/**
 * Created by debbie on 12/09/14.
 */
public class Xslt30TestSuiteDriver extends Xslt30TestSuiteDriverHE {



    public Xslt30TestSuiteDriver() {
        super();
    }


    public static void main(String[] args) throws Exception {
        if (args.length  < 2 || args[0].equals("-?")) {
            usage();
            System.exit(2);
        }

        new Xslt30TestSuiteDriver().go(args);
    }


    @Override
    public void go(String[] args) throws Exception {
        driverProc = new Processor(true);
        Licensor licensor = new Licensor();
        licensor.activate(driverProc);
        super.go(args);
    }

    /**
     * Return the appropriate tree model to use. This adds to the base set available
     * @param s  The name of the tree model required
     * @return  The tree model - null if model requested is unrecognised
     */
    protected TreeModel getTreeModel(String s) {
        TreeModel tree = super.getTreeModel(s);
        if (s.equalsIgnoreCase("jdom")) {
            tree = new JDOMObjectModel();
        } else if (s.equalsIgnoreCase("jdom2")) {
            tree = new JDOM2ObjectModel();
        } else if (s.equalsIgnoreCase("dom4j")) {
            tree = new DOM4JObjectModel();
        } else if (s.equalsIgnoreCase("xom")) {
            tree = new XOMObjectModel();
        } else if (s.equalsIgnoreCase("axiom")) {
            tree = new AxiomObjectModel();
        }
        return tree;
    }

    @Override
    protected void setLanguageVersion(XsltCompiler compiler, String spec) {
        compiler.setXsltLanguageVersion(spec.contains("XSLT30") || spec.contains("XSLT20+") ? "3.0" : "2.0");
    }

    @Override
    protected void initPatternOptimization(XsltCompiler compiler) {
       // compiler.getUnderlyingCompilerInfo().setPatternOptimization(new PatternOptimizationEE());
    }

    @Override
    public void prepareForSQL(Processor processor) {
        try {
            if (processor.getUnderlyingConfiguration() instanceof ProfessionalConfiguration) {
                //environment.processor.setConfigurationProperty(FeatureKeys.TRACE_EXTERNAL_FUNCTIONS, true);
                ((ProfessionalConfiguration) processor.getUnderlyingConfiguration())
                        .setExtensionElementNamespace("http://saxon.sf.net/sql", "net.sf.saxon.option.sql.SQLElementFactory");
            }
        } catch (XPathException e) {
            e.printStackTrace();
        }
    }

    /**
     * Ensure that a dependency is satisfied, first by checking whether Saxon supports
     * the requested feature, and if necessary by reconfiguring Saxon so that it does;
     * if configuration changes are made, then resetActions should be registered to
     * reverse the changes when the test is complete.
     *
     * @param dependency the dependency to be checked
     * @param env        the environment in which the test runs. The method may modify this
     *                   environment provided the changes are reversed for subsequent tests.
     * @return true if the test can proceed, false if the dependencies cannot be
     * satisfied.
     */

    @Override
    public boolean ensureDependencySatisfied(XdmNode dependency, Environment env) {
        String type = dependency.getNodeName().getLocalName();
        String value = dependency.getAttributeValue(new QName("value"));
        boolean inverse = "false".equals(dependency.getAttributeValue(new QName("satisfied")));
        if ("feature".equals(type)) {
            if ("schema_aware".equals(value)) {
                if (!treeModel.isSchemaAware() && !inverse) {
                    return false; // cannot use the selected tree model for schema-aware tests
                }
                if (env != null && env.processor.isSchemaAware() != !inverse) {
                    // Don't attempt to run non-SA tests with an SA processor; the presence of constructs like
                    // import schema will switch on schema-awareness.
                    return false;
                }
                if (env != null) {
                    env.resetActions.add(new Environment.ResetAction() {
                        @Override
                        public void reset(Environment env) {
                            env.xsltCompiler.setSchemaAware(false);
                        }
                    });
                    env.xsltCompiler.setSchemaAware(!inverse);
                    return true;
                } else {
                    return false;
                }
            } else if ("streaming".equals(value)) {
                if(!"EE".equals(env.processor.getSaxonEdition())) {
                    return inverse;
                }
                env.resetActions.add(new Environment.ResetAction() {
                    @Override
                    public void reset(Environment env) {
                        env.processor.setConfigurationProperty(FeatureKeys.STREAMABILITY, "off");
                        env.processor.setConfigurationProperty(FeatureKeys.STREAMING_FALLBACK, "off");
                    }
                });
                if (inverse) {
                    env.processor.setConfigurationProperty(FeatureKeys.STREAMABILITY, "off");
                    env.processor.setConfigurationProperty(FeatureKeys.STREAMING_FALLBACK, "off");
                } else {
                    switch (streaming) {
                        case 0: // -streaming:off
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMABILITY, "off");
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMING_FALLBACK, "on");
                            break;
                        case 1: // -streaming:std
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMABILITY, "standard");
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMING_FALLBACK, "off");
                            break;
                        case 2:
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMABILITY, "extended");
                            env.processor.setConfigurationProperty(FeatureKeys.STREAMING_FALLBACK, "off");
                            break;
                    }
                }
                return true;

            } else if ("XSD_1.1".equals(value)) {
                String requiredVersion = inverse ? "1.0" : "1.1";
                final String oldVersion = (String) env.processor.getConfigurationProperty(FeatureKeys.XSD_VERSION);
                if (env != null) {
                    env.resetActions.add(new Environment.ResetAction() {
                        @Override
                        public void reset(Environment env) {
                            env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, oldVersion);
                        }
                    });
                    env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, requiredVersion);
                    return true;
                } else {
                    return false;
                }

            }
//        } else if ("xsd-version".equals(type)) {
//            if ("1.1".equals(value)) {
//                if (env != null) {
//                    env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, inverse ? "1.0" : "1.1");
//                } else {
//                    return false;
//                }
//            } else if ("1.0".equals(value)) {
//                if (env != null) {
//                    env.processor.setConfigurationProperty(FeatureKeys.XSD_VERSION, inverse ? "1.1" : "1.0");
//                } else {
//                    return false;
//                }
//            }
//            return true;
        } else if ("sweep_and_posture".equals(type)) {
            if ("supports-sweep-and-posture-assessments".equals(value)) {
                return !inverse;
            }
            return !inverse;
        } else if ("xquery_invocation".equals(type)) {
            return !inverse;
        }
        return super.ensureDependencySatisfied(dependency, env);
    }

    /**
     * Run streamability tests
     */

    public void runStreamabilityTests(XPathCompiler xpc, XdmNode testCase) {
        try {
            String contextPostureStr = xpc.evaluateSingle("test/posture-and-sweep/@context-posture", testCase).getStringValue();
            Posture contextPosture = Posture.valueOf(contextPostureStr.toUpperCase());
            String resultPostureStr = xpc.evaluateSingle("result/assert-posture-and-sweep/@result-posture", testCase).getStringValue();
            Posture resultPosture = Posture.valueOf(resultPostureStr.toUpperCase());
            String resultSweepStr = xpc.evaluateSingle("result/assert-posture-and-sweep/@result-sweep", testCase).getStringValue();
            Sweep resultSweep = Sweep.valueOf(resultSweepStr.toUpperCase().replace('-', '_'));
            ContextItemStaticInfo info = new ContextItemStaticInfo(AnyItemType.getInstance(), true, contextPosture);
            XPathSelector tests = xpc.compile("test/posture-and-sweep/xpath").load();
            tests.setContextItem(testCase);
            Processor testProcessor = new Processor(true);
            XPathCompiler testCompiler = testProcessor.newXPathCompiler();
            testCompiler.declareNamespace("ex", "http://www.example.com");
            List<String> reasons = new ArrayList<String>();
            for (XdmItem test : tests) {
                try {
                    String testExpression = test.getStringValue();
                    XPathExecutable xExec = testCompiler.compile(testExpression);
                    Expression exp = xExec.getUnderlyingExpression().getInternalExpression();
                    PostureAndSweep ps = exp.getStreamability(false, info, reasons);
                    if (ps.getPosture().equals(resultPosture) && ps.getSweep().equals(resultSweep)) {
                        resultsDoc.writeTestcaseElement(testCase.getAttributeValue(new QName("name")), "pass", "");
                        successes++;
                    } else {
                        if (Literal.isEmptySequence(exp)) {
                            // no action: test bug 27075
                            // TODO: remove this when tests are fixed
                        } else if (testExpression.contains("namespace")) {
                            // no action: test bug 27078
                            // TODO: remove this when tests are fixed
                        } else {
                            resultsDoc.writeTestcaseElement(testCase.getAttributeValue(new QName("name")), "fail",
                                    testExpression + " is " + ps.toString());
                            failures++;
                        }
                    }
                } catch (SaxonApiException e) {
                    if (e.getMessage().equals("item() is not allowed in a path expression")) {
                        // TODO: remove this when tests are fixed
                        // no action: test bug 27074
                    } else {
                        resultsDoc.writeTestcaseElement(testCase.getAttributeValue(new QName("name")), "fail", e.getMessage());
                        failures++;
                    }
                }
            }

        } catch (SaxonApiException err) {
            resultsDoc.writeTestcaseElement(testCase.getAttributeValue(new QName("name")), "crashed", err.toString());
            err.printStackTrace();
        }
    }
}
