<?xml version="1.0" encoding="UTF-8"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
    xmlns:xs="http://www.w3.org/2001/XMLSchema"
    xmlns:ch="http://www.saxonica.com/ns/doc/changes"
    exclude-result-prefixes="xs ch"
    version="2.0">
    
    <!-- converts the changes.xml document to a regular article, so the standard article rendering code can be used -->
    
    <xsl:template match="/">
        <xsl:apply-templates select="*" mode="makepage"/>
    </xsl:template>
    
    <xsl:template match="ch:changes" mode="makepage">
        <xsl:message>Processing changes</xsl:message>
        <xsl:variable name="article">
            <article id="changes" title="Change History">
                <h1>Change History</h1>
                <p>In this section you will find a complete list of changes made to the Saxon product since release 7.0, organized in reverse
                order of releases.</p>
                <nav>
                    <ul/>
                </nav>
                <xsl:apply-templates mode="ch"/>
            </article> 
        </xsl:variable>
        
        <xsl:copy-of select="$article"/>
        
       
        <!--<xsl:apply-templates select="$article//(article|section)"  mode="makepage"/>-->
    </xsl:template>
    
    <xsl:template match="ch:categories" mode="ch"/>
    
    <xsl:template match="ch:release" mode="ch">
        <section id="v{@number}" title="Changes in Release {@number}">
            <h1>Changes in Release <xsl:value-of select="@number"/> (<xsl:value-of select="format-date(@date, '[D] [MNn] [Y]')"/>)</h1>
            <nav>
                <ul/>
            </nav>
            <xsl:apply-templates mode="ch"/>
        </section>
    </xsl:template>
    
    <xsl:template match="ch:category" mode="ch">
        <section id="{@name}" title="{/ch:changes/ch:categories/ch:cat[@name=current()/@name]/@title}">
            <h1><xsl:value-of select="/ch:changes/ch:categories/ch:cat[@name=current()/@name]/@title"/></h1>
            <xsl:apply-templates mode="ch"/>
        </section>
    </xsl:template>
    
    <xsl:template match="ch:change[@title]" mode="ch">
        <h2><xsl:value-of select="@title"/></h2>
        <xsl:apply-templates mode="ch"/>
    </xsl:template>
    
    <xsl:template match="ch:change" mode="ch">
        <xsl:apply-templates mode="ch"/>
    </xsl:template>
    
    <xsl:template match="*" mode="ch">
        <xsl:element name="{local-name()}">
            <xsl:copy-of select="@*"/>
            <xsl:apply-templates mode="ch"/>
        </xsl:element>
    </xsl:template>    
    
    <xsl:template match="ch:code[@java]" mode="ch">
        <a href="{@java}" class="javalink">
            <xsl:copy-of select="@* except @java|@class"/>
            <xsl:apply-templates/>
        </a>
    </xsl:template>
    
    
    
</xsl:stylesheet>